# This is where you can add stuff to the admin view
from django.contrib import admin
from django_teams.models import Team, TeamStatus, Ownership
from django.contrib.auth.models import Group, User

try:
    admin.site.unregister(Group)
    admin.site.unregister(User)
except:  # noqa: E722
    pass

@admin.register(Group)
class GroupAdmin(admin.ModelAdmin):

    show_full_result_count = False


@admin.register(User)
class UserAdmin(admin.ModelAdmin):

    show_full_result_count = False


@admin.register(Team)
class TeamAdmin(admin.ModelAdmin):

    show_full_result_count = False

    def get_queryset(self, request):
        qs = super(TeamAdmin, self).get_queryset(request)
        return qs.select_related()


@admin.register(Ownership)
class OwnershipAdmin(admin.ModelAdmin):

    show_full_result_count = False

    def get_queryset(self, request):
        qs = super(OwnershipAdmin, self).get_queryset(request)
        return qs.select_related()


@admin.register(TeamStatus)
class TeamStatusAdmin(admin.ModelAdmin):

    show_full_result_count = False

    def get_queryset(self, request):
        qs = super(TeamStatusAdmin, self).get_queryset(request)
        return qs.select_related('user', 'team').filter(user=request.user)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-06 20:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_teams', '0002_ownership_approved'),
    ]

    operations = [
        migrations.AddField(
            model_name='team',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='team',
            name='users',
            field=models.ManyToManyField(blank=True, null=True, related_name='team_member', through='django_teams.TeamStatus', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='teamstatus',
            name='comment',
            field=models.CharField(default='', max_length=255),
        ),
        migrations.AlterField(
            model_name='teamstatus',
            name='role',
            field=models.IntegerField(choices=[(1, 'Requesting Access'), (10, 'Team Member'), (20, 'Team Leader')]),
        ),
    ]



from django.core.exceptions import PermissionDenied
# This is where your views go :)
from django.http import HttpResponseRedirect
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.db.models import Count, F
from django_teams.models import Team, TeamStatus, Ownership
from django_teams.forms import (TeamEditForm,
                                TeamStatusCreateForm,
                                action_formset)
from django.db.models import Case, When
from django.db import models
from django.contrib.contenttypes.models import ContentType


def loadGenericKeyRelations(queryset):
    distinct_contents = queryset.values_list('content_type', flat=True).distinct()
    object_items = []
    for object in distinct_contents:
        content_type = ContentType.objects.get(id=object).model_class()
        set = queryset.filter(content_type=object).values()
        objects = content_type.objects.filter(pk__in=[object['object_id'] for object in set])
        for relation in content_type._meta.get_fields():
            if relation.get_internal_type() is 'ForeignKey':
                objects.select_related(relation.name)
        object_items.append(objects.all())
    return object_items


class TeamListView(ListView):
    model = Team

    def render_to_response(self, context, **response_kwargs):
        queryset = Team.objects.all().annotate(member_count=Count('users'))
        queryset = queryset.annotate(owner=Case(When(teamstatus__role=20, then=F('users__username')), default=None))
        if not self.request.user.is_anonymous():
            queryset = queryset.annotate(role=Case(When(teamstatus__user=self.request.user,
                                         then=F('teamstatus__role')), default=0, outputfield=models.IntegerField()))
            queryset = queryset.order_by('-role')
        else:
            queryset = queryset.order_by('-id')
        team_names = []
        team_list = []
        # combine teams with the same name
        for q in queryset:
            if q.name not in team_names:
                team_names.append(q.name)
                tmp = {'name': q.name, 'id': q.id, 'pk': q.pk, 'description': q.description,
                       'member_count': q.member_count, 'owner': q.owner}
                try:
                    tmp['role'] = q.role
                except Exception:
                    tmp['role'] = None
                team_list.append(tmp)
            else:
                t = team_list[team_names.index(q.name)]
                t['member_count'] += q.member_count
                if q.owner is not None:
                    t['owner'] = q.owner
        return super(TeamListView, self).render_to_response({'list': team_list}, **response_kwargs)


class UserTeamListView(ListView):
    template_name = 'django_teams/user_team_list.html'

    def get_queryset(self):
        statuses = TeamStatus.objects.select_related('user').filter(user=self.request.user,
                                                                    role=20).values_list('team', flat=True)
        return statuses


class TeamCreateView(CreateView):
    model = Team
    template_name = 'django_teams/team_create.html'
    fields = ['name', 'description', 'private']

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        return super(TeamCreateView, self).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        ret = super(TeamCreateView, self).form_valid(form)
        self.object.add_user(self.request.user, team_role=20)
        return ret


class TeamDetailView(DetailView):
    model = Team

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        return super(TeamDetailView, self).dispatch(request, *args, **kwargs)

    def get_object(self, queryset=None):
        object = super(TeamDetailView, self).get_object(queryset)
        if object.private and self.request.user not in object.users.filter(teamstatus__role__gte=10):
            raise PermissionDenied()
        return object

    def render_to_response(self, context, **response_kwargs):
        team = self.object
        # context['owner'] = team.users.filter(teamstatus__role=20)
        # context['members'] = team.users.filter(teamstatus__role=10)
        context['owners'] = []
        context['members'] = []
        statuses = TeamStatus.objects.select_related('user', 'team').filter(team=team)
        for s in statuses:
            if s.role == 10:
                context['members'].append(s.user)
            elif s.role == 20:
                context['owners'].append(s.user)
        owned = Ownership.objects.filter(team=team, approved=True)
        context['approved_objects_types'] = loadGenericKeyRelations(owned)
        return super(TeamDetailView, self).render_to_response(context, **response_kwargs)


class TeamInfoEditView(UpdateView):
    model = Team
    fields = ['name', 'description', 'private']
    template_name = 'django_teams/teaminfo_form.html'

    def get_object(self, queryset=None):
        object = super(TeamInfoEditView, self).get_object(queryset)
        if self.request.user not in object.users.filter(teamstatus__role__gte=20):
            raise PermissionDenied()
        return object


class TeamEditView(UpdateView):
    model = Team
    form_class = TeamEditForm

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        return super(TeamEditView, self).dispatch(request, *args, **kwargs)

    def get_object(self, queryset=None):
        object = super(TeamEditView, self).get_object(queryset)

        # User must be admin of the object to get into this view
        if self.request.user not in object.users.filter(teamstatus__role__gte=20):
            raise PermissionDenied()
        return object

    def get_form_class(self):
        # get forms for team leaders, team members, team requests
        ret = []
        users = self.object.users
        ret += [action_formset(prefix_name='teachers', qset=users.filter(teamstatus__role=20),
                               actions=('---', 'Demote', 'Remove'))]
        ret += [action_formset(prefix_name='students', qset=users.filter(teamstatus__role=10),
                               actions=('---', 'Promote', 'Remove'))]
        ret += [action_formset(prefix_name='member-requests', qset=users.filter(teamstatus__role=1),
                               actions=('---', 'Approve', 'Reject'))]
        owned_objects = Ownership.objects.filter(team=self.object)
        approved = loadGenericKeyRelations(owned_objects.filter(approved=True))
        for set in approved:
            if set:
                prefix_name = 'approved-' + str(set.model.__name__)
                ret += [action_formset(prefix_name=prefix_name, qset=set, actions=('---', 'Remove'), link=True)]
        pending_approval = loadGenericKeyRelations(owned_objects.filter(approved=False))
        for set in pending_approval:
            if set:
                prefix_name = str(set.model.__name__) + 's-pending-approval'
                ret += [action_formset(prefix_name=prefix_name, qset=set,
                                       actions=('---', 'Approve', 'Remove'), link=True)]
        return ret

    def get_form(self, form_class=TeamEditForm):
        kwargs = self.get_form_kwargs()
        form_class = self.get_form_class()
        if 'data' in kwargs:
            ret = [form_class[num](kwargs['data'],
                                   prefix=form_class[num].name) for num in range(len(form_class))]
        else:
            ret = [form_class[num](prefix=form_class[num].name) for num in range(len(form_class))]

        return ret

    def post(self, request, *args, **kwargs):
        """
        Handles POST requests, instantiating a form instance with the passed
        POST variables and then checked for validity.
        """
        self.object = get_object_or_404(Team, pk=kwargs['pk'])
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        for f in form:
            if not f.is_valid():
                return self.form_invalid(form)

        # Go through each form and perform the action
        # Owners
        owner_action = form[0].cleaned_data['action']
        owner_items = form[0].cleaned_data['items']
        statuses = TeamStatus.objects.filter(team=self.object, user__in=[object for object in owner_items])
        if owner_action == 'Demote':
            for status in statuses:
                status.role = 10
                status.save()
        if owner_action == 'Remove':
            for status in statuses:
                status.delete()

        # Members
        member_action = form[1].cleaned_data['action']
        member_items = form[1].cleaned_data['items']
        statuses = TeamStatus.objects.filter(team=self.object, user__in=[object for object in member_items])
        if member_action == 'Promote':
            for status in statuses:
                status.role = 20
                status.save()
        if member_action == 'Remove':
            for status in statuses:
                status.delete()

        # Member Requests
        request_action = form[2].cleaned_data['action']
        request_items = form[2].cleaned_data['items']
        statuses = TeamStatus.objects.filter(team=self.object, user__in=[object for object in request_items])
        if request_action == 'Approve':
            for status in statuses:
                status.role = 10
                status.save()
        if request_action == 'Revoke':
            for status in statuses:
                status.delete()

        for num in range(3, len(form)):
            current_action = form[num].cleaned_data['action']
            current_items = form[num].cleaned_data['items']
            objects = []
            if current_items:
                content_type = ContentType.objects.get_for_model(current_items[0])
                objects = Ownership.objects.filter(content_type=content_type,
                                                   team=self.object,
                                                   object_id__in=[object.id for object in current_items])
            if current_action == 'Approve':
                for object in objects:
                    object.approved = True
                    object.save()
            if current_action == 'Remove':
                for object in objects:
                    object.delete()

        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return reverse('team-edit', kwargs={'pk': self.object.pk})


class TeamStatusCreateView(CreateView):
    model = TeamStatus
    form_class = TeamStatusCreateForm

    def get_success_url(self):
        return reverse('team-list')

    def dispatch(self, request, *args, **kwargs):
        self.team = Team.objects.get(pk=kwargs['team_pk'])
        self.request = request
        return super(TeamStatusCreateView, self).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        form.instance.team = self.team
        form.instance.user = self.request.user
        form.instance.role = 1
        return super(TeamStatusCreateView, self).form_valid(form)

    def render_to_response(self, context, **response_kwargs):
        context['team'] = self.team
        return super(TeamStatusCreateView, self).render_to_response(context, **response_kwargs)



from django import template
from django.contrib.contenttypes.models import ContentType


def get_user_status(team, user):
    ts = team.get_user_status(user)
    return ts


def get_owned_objects(team, model):
    return team.owned_objects(model)


def get_approved_objects(team, model):
    return team.approved_objects_of_model(model)


def verbose_name(object):
    return object._meta.verbose_name


def load_fragment_template(object):
    ct = ContentType.objects.get_for_model(object)
    template_name = 'django_teams/fragments/%s/%s.html' % (ct.app_label, ct.model)
    try:
        template.loader.get_template(template_name)
        return template_name
    except template.TemplateDoesNotExist:
        return "django_teams/fragments/default.html"


def get(object, name):
    return getattr(object, name)


register = template.Library()
register.filter(get_user_status)
register.filter(get_owned_objects)
register.filter(get_approved_objects)
register.filter(verbose_name)
register.filter(load_fragment_template)
register.filter(get)



from django.conf.urls import include, url
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.views import static
from django_teams.views import (TeamListView, UserTeamListView,
                                TeamCreateView, TeamDetailView,
                                TeamInfoEditView, TeamEditView,
                                TeamStatusCreateView)

from django.contrib import admin
admin.autodiscover()


urlpatterns = [
    url(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    url(r'^admin/', include(admin.site.urls)),
    url(r'^teams/$', TeamListView.as_view(), name='team-list'),
    url(r'^my-teams/$', login_required(UserTeamListView.as_view()), name='user-team-list'),
    url(r'^teams/create$', login_required(TeamCreateView.as_view()), name='team-create'),
    url(r'^teams/(?P<team_pk>\d+)/invite$', login_required(TeamStatusCreateView.as_view()), name='teamstatus-create'),
    url(r'^teams/(?P<pk>\d+)/$', TeamDetailView.as_view(), name='team-detail'),
    url(r'^teams/(?P<pk>\d+)/edit$', TeamEditView.as_view(), name='team-edit'),
    url(r'^teams/(?P<pk>\d+)/info/edit$', TeamInfoEditView.as_view(), name='team-info-edit'),
]

if settings.DEBUG:
    urlpatterns += [
        url(r'^media/(?P<path>.*)$', static.serve,
            {'document_root': settings.MEDIA_ROOT, }), ]



from time import sleep
from django.test import TestCase
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.contenttypes.models import ContentType
from django_teams.models import Team, TeamStatus, Ownership


class TeamTests(TestCase):
    fixtures = ['test_data.json']

    def test_can_create_new_team(self):
        original_count = Team.objects.all().count()
        team = Team(name="Team Awesome")
        team.save()
        sleep(1)

        self.assertEqual(Team.objects.all().count(), original_count + 1)

    def test_can_add_user_to_team(self):
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)

        original_count = team.users.all().count()
        team.add_user(user)
        sleep(1)

        self.assertEqual(team.users.all().count(), original_count + 1)

    def test_can_approve_user(self):
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)

        original_count = team.users.all().count()
        team.add_user(user)
        sleep(1)

        self.assertEqual(team.users.all().count(), original_count + 1)

        ts = TeamStatus.objects.filter(user=user, team=team).reverse()[0]

        self.assertEqual(ts.role, 1)

        team.approve_user(user)

        ts = TeamStatus.objects.filter(user=user, team=team).reverse()[0]

        self.assertEqual(ts.role, 10)

    def test_cant_approve_user_with_no_invite(self):
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)

        ts_count = TeamStatus.objects.filter(user=user, team=team).count()

        self.assertEqual(ts_count, 0)

        # Should throw error DNE
        try:
            team.approve_user(user)
        except ObjectDoesNotExist:
            pass

        ts_count = TeamStatus.objects.filter(user=user, team=team).count()

        self.assertEqual(ts_count, 0)

    def test_can_get_team_owners(self):
        team = Team(name="Team Awesome")
        team.save()
        user = User.objects.get(pk=1)

        team.add_user(user, team_role=20)
        owners = team.users.filter(teamstatus__role=20)
        self.assertIn(user, owners)

    def test_can_get_owned_objects(self):
        team = Team(name="Team Awesome")
        team.save()
        user = User.objects.get(pk=1)

        Ownership.grant_ownership(team, user)

        contenttype = ContentType.objects.get_for_model(User)
        owned_objects = []
        ownership_set = Ownership.objects.select_related('team',
                                                         'content_type').filter(team=team, content_type=contenttype)
        for ownership in ownership_set.iterator():
            owned_objects += [ownership.content_object]

        self.assertIn(user, owned_objects)

    def test_can_get_list_of_object_types(self):
        team = Team(name="Team Awesome")
        team.save()
        user = User.objects.get(pk=1)

        Ownership.grant_ownership(team, user)

        owned_object_types = []
        ownership_set = Ownership.objects.select_related('team').filter(team=team)
        for ownership in ownership_set.iterator():
            if ownership.content_type.model_class() not in owned_object_types:
                owned_object_types += [ownership.content_type.model_class()]

        self.assertIn(User, owned_object_types)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-17 18:45
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_teams', '0004_auto_20171117_1818'),
    ]

    operations = [
        migrations.AlterField(
            model_name='team',
            name='users',
            field=models.ManyToManyField(blank=True, related_name='team_member', through='django_teams.TeamStatus', to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.forms import ModelForm
from django import forms
from django_teams.models import Team, TeamStatus


class TeamCreateForm(ModelForm):
    class Meta:
        model = Team
        fields = ['name', 'description', 'private']


class TeamEditForm(ModelForm):
    """This form is very complicated;
    it consists of a hash containing :
        An array of pending requests
        An array of team leaders
        An array of team members
    The form should allow the team leaders to perform the following
    actions on any number of elements from each array, at the same time:
        Pending requests;
            approve
            deny
        Team Leaders
            Demote
            Remove
        Team Members
            Promote
            Remove
    """

    class Meta:
        model = Team
        fields = ['users', 'name', 'description', 'private']


class TeamStatusCreateForm(ModelForm):
    class Meta:
        model = TeamStatus
        fields = ['comment']


def action_formset(prefix_name, qset, actions, link=False):
    """A form factory which returns a form which allows the user to pick a specific action to
    perform on a chosen subset of items from a queryset.
    """
    class _ActionForm(forms.Form):
        name = prefix_name
        queryset = qset
        items = forms.ModelMultipleChoiceField(queryset=qset, required=False)
        action = forms.ChoiceField(choices=zip(actions, actions), required=False)

    return _ActionForm



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-17 18:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_teams', '0003_auto_20160906_2058'),
    ]

    operations = [
        migrations.AlterField(
            model_name='teamstatus',
            name='comment',
            field=models.CharField(blank=True, default=b'', max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Ownership',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('private', models.BooleanField(default=False)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TeamStatus',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('comment', models.CharField(default=b'', max_length=255)),
                ('role', models.IntegerField(choices=[(1, b'Requesting Access'), (10, b'Team Member'), (20, b'Team Leader')])),
                ('team', models.ForeignKey(to='django_teams.Team')),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='team',
            name='users',
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL, null=True, through='django_teams.TeamStatus', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='ownership',
            name='team',
            field=models.ForeignKey(to='django_teams.Team'),
            preserve_default=True,
        ),
    ]



# This is where the models go!
from django.db import models
from django.core.urlresolvers import reverse
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey

# Using the user: user = models.ForeignKey(settings.AUTH_USER_MODEL)
CurrentUser = None
CurrentTeam = None


class Team(models.Model):
    users = models.ManyToManyField(settings.AUTH_USER_MODEL,
                                   null=True,
                                   blank=True,
                                   through='django_teams.TeamStatus',
                                   related_name='team_member')
    name = models.CharField(max_length=255)
    private = models.BooleanField(default=False)
    description = models.TextField(null=True, blank=True)

    def get_absolute_url(self):
        return reverse('team-detail', kwargs={'pk': self.pk})

    def __unicode__(self):
        return self.name

    def add_user(self, user, team_role=1):
        TeamStatus(user=user, team=self, role=team_role).save()

    def approve_user(self, user):
        ts = TeamStatus.objects.get(user=user, team=self)
        if ts.role == 1:
            ts.role = 10
            ts.save()

    def approved_objects(self):
        return Ownership.objects.select_related('team').filter(team=self, approved=True)

    @staticmethod
    def get_current_team():
        if CurrentTeam is not None:
            return CurrentTeam
        return None


class TeamStatus(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL)
    team = models.ForeignKey('django_teams.Team')
    comment = models.CharField(max_length=255, default='', null=True, blank=True)

    TEAM_ROLES = (
        (1, 'Requesting Access'),
        (10, 'Team Member'),
        (20, 'Team Leader'),
    )

    role = models.IntegerField(choices=TEAM_ROLES)

    def approve(self):
        self.role = 10
        self.save()

    def __unicode__(self):
        return "%s requesting to join %s" % (self.user.__unicode__(), self.team.__unicode__())


class Ownership(models.Model):
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    approved = models.BooleanField(default=False)
    team = models.ForeignKey('django_teams.Team')

    @staticmethod
    def check_permission(item):
        content_type = ContentType.objects.get_for_model(item)
        res = Ownership.objects.filter(team=Team.get_current_team(), content_type=content_type, object_id=item.id)

        return len(res) > 0

    @staticmethod
    def grant_ownership(team, item):
        content_type = ContentType.objects.get_for_model(item)

        res = Ownership.objects.get_or_create(team=team, content_type=content_type, object_id=item.id)
        if res[1]:
            res[0].save()



def required(wrapping_functions, patterns_rslt):
    '''
    Used to require 1..n decorators in any view returned by a url tree

    Usage:
      urlpatterns = required(func,patterns(...))
      urlpatterns = required((func,func,func),patterns(...))

    Note:
      Use functools.partial to pass keyword params to the required
      decorators. If you need to pass args you will have to write a
      wrapper function.

    Example:
      from functools import partial

      urlpatterns = required(
          partial(login_required,login_url='/accounts/login/'),
          patterns(...)
      )
    '''
    if not hasattr(wrapping_functions, '__iter__'):
        wrapping_functions = (wrapping_functions, )
    return [
        _wrap_instance__resolve(wrapping_functions, instance)
        for instance in patterns_rslt
    ]


def _wrap_instance__resolve(wrapping_functions, instance):
    if not hasattr(instance, 'resolve'):
        return instance
    resolve = getattr(instance, 'resolve')

    def _wrap_func_in_returned_resolver_match(*args, **kwargs):
        rslt = resolve(*args, **kwargs)
        if not hasattr(rslt, 'func'):
            return rslt
        f = getattr(rslt, 'func')
        for _f in reversed(wrapping_functions):
            # @decorate the function from inner to outter
            f = _f(f)
        setattr(rslt, 'func', f)
        return rslt
    setattr(instance, 'resolve', _wrap_func_in_returned_resolver_match)
    return instance


def get_related_managers(self):
    managers = []
    for related_object in self._meta.get_all_related_objects():
        try:
            managers.append(getattr(self, related_object.get_accessor_name()))
        except:
            pass
    return managers



from django.test import TestCase
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django_teams.models import Team


class ListTeamsTests(TestCase):
    fixtures = ['test_data.json']

    def test_can_get_route(self):
        self.assertTrue(reverse('team-list') is not None)

    def test_can_load_page(self):
        response = self.client.get(reverse('team-list'))

        self.assertEqual(response.status_code, 200)

    def test_page_contains_all_teams(self):
        # Team(name="Team Awesome").save()
        # Team(name="Team Silly").save()
        # Team(name="Hamburger").save()

        Team(name='team1').save()
        Team(name='team2').save()
        Team(name='team3').save()
        self.assertTrue(Team.objects.all().count() > 0)

        response = self.client.get(reverse('team-list'))

        for team in Team.objects.all():
            self.assertContains(response, str(team))

    def test_page_contains_links_to_teams(self):
        self.test_page_contains_all_teams()

        response = self.client.get(reverse('team-list'))

        for team in Team.objects.all():
            self.assertContains(response, reverse('team-detail', kwargs={'pk': team.pk}))


class DetailTeamsTests(TestCase):
    fixtures = ['test_data.json']

    def test_can_get_route(self):
        self.assertTrue(reverse('team-detail', kwargs={'pk': 1}) is not None)

    def test_can_load_page(self):
        response = self.client.get(reverse('team-detail', kwargs={'pk': 1}))

        self.assertEqual(response.status_code, 200)

    def test_contains_team_name(self):
        team = Team.objects.get(pk=1)
        response = self.client.get(reverse('team-detail', kwargs={'pk': team.pk}))

        self.assertContains(response, str(team))

    def test_contains_list_of_owners(self):
        team = Team.objects.get(pk=1)
        response = self.client.get(reverse('team-detail', kwargs={'pk': team.pk}))
        owners = team.users.filter(teamstatus__role=20)
        for leader in owners:
            self.assertContains(response, str(leader))

    def test_private_team_is_not_open_to_public(self):
        team = Team(name="Team Awesome", private=True)
        team.save()

        response = self.client.get(reverse('team-detail', kwargs={'pk': team.pk}))

        self.assertEqual(response.status_code, 403)


class EditTeamsTests(TestCase):
    fixtures = ['test_data.json']

    def test_can_get_route(self):
        self.assertTrue(reverse('team-edit', kwargs={'pk': 1}) is not None)

    def test_can_tell_admin_page(self):
        """Verify text asserting that this is the admin page
        """
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)
        team.add_user(user, team_role=20)

        self.client.login(username='test', password='test')
        response = self.client.get(reverse('team-edit', kwargs={'pk': team.pk}))
        self.client.logout()

        self.assertContains(response, str(team))
        self.assertContains(response, "admin")

    def test_non_leader_cant_access_page(self):
        """Non-admin users should not be able to access this page at all
        """
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)
        team.add_user(user, team_role=10)

        self.client.login(username='test', password='test')

        response = self.client.get(reverse('team-edit', kwargs={'pk': team.pk}))

        self.assertEqual(response.status_code, 403)

        self.client.logout()

    def test_can_post_page(self):
        """Verify that we can post changes to an admin page;
        for example, approve users
        """
        pass

    def test_can_tell_admin_info_page(self):
        """Verify that we can access the update info page
        """
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)
        team.add_user(user, team_role=20)

        self.client.login(username='test', password='test')
        self.client.get(reverse('team-info-edit', kwargs={'pk': team.pk}))
        self.client.logout()

    def test_non_leader_cant_access_info_page(self):
        """Non-admin users should not be able to access this page at all
        """
        team = Team(name="Team Awesome")
        team.save()

        user = User.objects.get(pk=1)
        team.add_user(user, team_role=10)

        self.client.login(username='test', password='test')

        response = self.client.get(reverse('team-info-edit', kwargs={'pk': team.pk}))

        self.assertEqual(response.status_code, 403)

        self.client.logout()
